#include <stdio.h>
#include <string.h>
#include <netdb.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include "net.h"

#ifndef CHECK
#define CHECK 0
#endif

#ifdef sun
extern int errno;
#endif /* sun */

#ifndef SOMAXCONN
#define SOMAXCONN 5		/* min(SunOS, OSF1, BSD/OS) */
#endif

void so_reuseaddr(int sfd)
{
#ifdef SO_REUSEADDR
    int on = 1;
    setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, (caddr_t)&on, sizeof(on));
#endif /* SO_REUSEADDR */
}

int open_server_socket(unsigned short port)
{
    int sfd;
    struct sockaddr_in server;

    if((sfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	return -1;

    pack_sockaddr_in(&server, port, htonl(INADDR_ANY));
    so_reuseaddr(sfd);
    if(bind_in(sfd, &server) < 0)
    {
	close(sfd);
	return -2;
    }

    /* Set it up to wait for connections. */
    if(listen(sfd, SOMAXCONN) < 0)
    {
	close(sfd);
	return -3;
    }

    return sfd;
}

int connect_to_server(const char* hostname, unsigned short tcp_port)
{
    int fd;
    struct sockaddr_in in;

    if((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
	perror("socket");
	return -1;
    }

    pack_sockaddr_in(&in, tcp_port, name2addr(hostname));
    if(connect_in(fd, &in) < 0)
    {
	close(fd);
	return -2;
    }

    return fd;
}

int nread(int fd, void* buff, unsigned long nbytes)
{
    int n, count, rest;

    count = 0;
    rest  = nbytes;
    while((n = read(fd, (char *)buff + count, rest)) <= 0)
    {
	rest  -= n;
	count += n;
	if(rest <= 0)
	    break;
    }

    if(n < 0)
	return -1;
    return nbytes;
}

int nwrite(int fd, void* buff, unsigned long nbytes)
{
    int n, count, rest;

    count = 0;
    rest  = nbytes;
    while((n = write(fd, (char *)buff + count, rest)) <= 0)
    {
	rest  -= n;
	count += n;
	if(rest <= 0)
	    break;
    }

    if(n < 0)
	return -1;
    return nbytes;
}

void pack_sockaddr_in(struct sockaddr_in *s,
		      unsigned short port,
		      unsigned int   iaddr)
{
    memset(s, 0, sizeof(struct sockaddr_in));
    s->sin_port = htons(port);
    s->sin_addr.s_addr = iaddr;
    s->sin_family = AF_INET;
}

unsigned int name2addr(const char* name)
{
    unsigned int addr;
    struct hostent *hp;

    if((addr = inet_addr(name)) != INADDR_NONE)
        return addr;

    if((hp = gethostbyname(name)) == NULL)
        return INADDR_NONE;
    memcpy(&addr, hp->h_addr, hp->h_length);
    return addr;
}

int recvfrom_in(int sfd, char* packet, int maxsize, struct sockaddr_in* from)
{
    int len, n;
#if CHECK
    extern void xdump();

#endif
    do
    {
	len = sizeof(struct sockaddr_in);
	memset(from, 0, len);
	n = recvfrom(sfd, packet, maxsize, 0, (struct sockaddr *)from, &len);
    } while(n < 0 && errno == EINTR);

#if CHECK
    printf("##%d: recv: %s:%d:", getpid(),
	   inet_ntoa(from->sin_addr), ntohs(from->sin_port));
    xdump(packet, n); putchar('\n');
#if 0
{
    struct sockaddr_in my;
    int t;
    getsockname(sfd, (struct sockaddr *)&my, &t);
    printf("## %u %u\n", my.sin_addr.s_addr, my.sin_port);
}
#endif
#endif


    return n;
}
    
int sendto_in(int sfd, char* packet, int size, struct sockaddr_in* to)
{
    int n;
#if CHECK
    extern void xdump();
    extern unsigned int rand8();
    struct sockaddr_in from;

    getsockname(sfd, (struct sockaddr *)&from, &n);
    printf("##%d:send:", getpid()); xdump(packet, size); putchar('\n');
    printf("##%d:from: %s:%d\n",
	   getpid(), inet_ntoa(from.sin_addr), ntohs(from.sin_port));
    printf("##%d:to: %s:%d\n",
	   getpid(), inet_ntoa(to->sin_addr), ntohs(to->sin_port));



#if 0
    if(rand8() < 127)
    {
	printf("## Packet lost\n");
	return n; /* Packet lost intentionally */
    }
#endif
#endif

    do
	n = sendto(sfd, packet, size, 0,
		   (struct sockaddr *)to, sizeof(struct sockaddr_in));
    while(n < 0 && (errno == ENOBUFS || errno == EINTR));
    return n;
}

int bind_in(int sfd, struct sockaddr_in* addr)
{
    return bind(sfd, (struct sockaddr *)addr, sizeof(struct sockaddr_in));
}

int simple_send_udp(char* packet, int size,
		    unsigned short port, unsigned int ipaddr)
{
    int sfd;
    struct sockaddr_in addr;

    if((sfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
    {
	perror("socket");
	return -1;
    }
    pack_sockaddr_in(&addr, port, ipaddr);
    if(sendto_in(sfd, packet, size, &addr) < 0)
    {
	perror("sendto");
	close(sfd);
	return -1;
    }
    close(sfd);
    return 0;
}

int connect_in(int fd, struct sockaddr_in* addr)
{
    return connect(fd, (struct sockaddr *)addr, sizeof(struct sockaddr_in));
}

int accept_and_fork(int sfd, struct sockaddr_in* from)
{
    int fd, pid, len;

    while(1)
    {
	if(from == NULL)
	    fd = accept(sfd, NULL, NULL);
	else
	{
	    len = sizeof(struct sockaddr_in);
	    memset(from, 0, len);
	    fd = accept(sfd, (struct sockaddr *)from, &len);
	}
	if(fd < 0)
        {
            if(errno == EINTR)
                continue;
            perror("accept");
	    sleep(3);
            continue;
        }

        if((pid = fork()) < 0)
        {
            perror("fork");
            sleep(3);
            continue;
        }

        if(pid == 0)            /* child */
	{
	    close(sfd);
	    return fd;
	}
        close(fd);
    }
}



#if 0
void main(void)
{
    int sfd, fd, n;
    char buff[BUFSIZ];

    sfd = open_server_socket(12345);
    fd = accept_and_fork(sfd, NULL);
    while((n = read(fd, buff, BUFSIZ)) > 0)
    {
	buff[n] = '\0';
	printf("%d: %s\n", getpid(), buff);
    }
    close(fd);
}
#endif

struct hostent *local_hostent(void)
{
    char host[64];
    if(gethostname(host, 64) == NULL)
	return NULL;
    return gethostbyname(host);
}
